import argparse
import os

from jenesis.contracts.detect import parse_contract
from jenesis.config import Config


def run(args: argparse.Namespace):

    project_root = os.path.abspath(os.getcwd())
    contract_root = os.path.join(project_root, "contracts", args.contract)

    # check that we are actually running the command from the project root
    if not os.path.exists(os.path.join(project_root, "jenesis.toml")):
        print("Please run command from project root")
        return

    # check if the contract exists
    if not os.path.exists(contract_root):
        print(f'Contract "{args.contract}" does not exist')
        return

    cfg = Config.load(project_root)

    profile_name = cfg.get_default_profile()

    selected_contract = parse_contract(project_root , args.contract)


    for (profile_name, profile) in cfg.profiles.items():
        network_name = profile.network.name
        for deployment in args.deployments:
            Config.update_project(os.getcwd(), profile_name, network_name, selected_contract, deployment)


def add_deployment_command(parser):
    deployment_cmd = parser.add_parser("deployment")
    deployment_cmd.add_argument("contract", help="The name of the contract that the deployment points to")
    deployment_cmd.add_argument('deployments', nargs='+', help='contract deployments',)
    deployment_cmd.set_defaults(handler=run)



import argparse
import sys
from typing import Tuple

from jenesis.cmd.add import add_add_command
from jenesis.cmd.compile import add_compile_command
from jenesis.cmd.init import add_init_command
from jenesis.cmd.new import add_new_command
from jenesis.cmd.attach import add_attach_command
from jenesis.cmd.update import add_update_command
from jenesis.cmd.deploy import add_deploy_command
from jenesis.cmd.run import add_run_command
from jenesis.cmd.shell import add_shell_command
from jenesis.cmd.keys import add_keys_command
from jenesis.cmd.network import add_network_command


def _parse_commandline() -> Tuple[argparse.ArgumentParser, argparse.Namespace]:
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()

    add_deploy_command(subparsers)
    add_run_command(subparsers)
    add_shell_command(subparsers)
    add_keys_command(subparsers)
    add_init_command(subparsers)
    add_new_command(subparsers)
    add_compile_command(subparsers)
    add_add_command(subparsers)
    add_attach_command(subparsers)
    add_update_command(subparsers)
    add_network_command(subparsers)

    return parser, parser.parse_args()


def main():
    parser, args = _parse_commandline()

    exit_code = 1
    if hasattr(args, 'handler'):
        result = args.handler(args)
        if isinstance(result, bool) and result:
            exit_code = 0
        elif isinstance(result, int):
            exit_code = result
        elif result is None:
            exit_code = 0

    else:
        parser.print_usage()

    if exit_code != 0:
        sys.exit(exit_code)



import argparse
import os
import sys

from cosmpy.aerial.client import LedgerClient
from cosmpy.aerial.faucet import FaucetApi
from cosmpy.aerial.wallet import LocalWallet
from cosmpy.crypto.keypairs import PrivateKey
from ptpython import embed
from jenesis.config import Config, Profile
from jenesis.contracts.detect import detect_contracts
from jenesis.contracts.monkey import make_contract
from jenesis.contracts.observer import DeploymentUpdater
from jenesis.keyring import query_keychain_items, query_keychain_item
from jenesis.network import network_context


PROJECT_PATH = os.getcwd()


def get_profile(cfg: Config, args: argparse.Namespace) -> Profile:

    # check that we are actually running the command from the project root
    if not os.path.exists(os.path.join(PROJECT_PATH, "jenesis.toml")):
        raise RuntimeError("Please run command from project root or create project first")

    if args.profile is not None and args.profile not in cfg.profiles:
        print(f'Invalid profile name. Expected one of {",".join(cfg.profiles.keys())}')
        sys.exit(1)

    profile_name = args.profile or cfg.get_default_profile()

    return cfg.profiles.get(profile_name)


def load_shell_globals(cfg: Config, selected_profile: Profile) -> dict:

    shell_globals = {}
    contract_instances = {}

    contracts = {contract.name: contract for contract in detect_contracts(PROJECT_PATH)}

    deployments = selected_profile.deployments

    if selected_profile is not None:

        # build the ledger client
        client = LedgerClient(selected_profile.network)

        print(f'Network: {selected_profile.network.name}')

        print('Detecting contracts...')

        for (deployment_name, deployment) in deployments.items():
            contract = contracts[deployment.contract]
            print("C", deployment_name)

            # skip contracts that we have not compiled
            if contract.digest() is None:
                continue

            monkey = make_contract(
                contract,
                client,
                address=deployment.address,
                code_id=deployment.code_id,
                observer=DeploymentUpdater(
                    cfg,
                    PROJECT_PATH,
                    selected_profile.name,
                    deployment_name,
                ),
                init_args=deployment.init,
            )

            contract_instances[deployment_name] = monkey

        print('Detecting contracts...complete')

        shell_globals["ledger"] = LedgerClient(selected_profile.network)
        if selected_profile.network.faucet_url is not None:
            shell_globals["faucet"] = FaucetApi(selected_profile.network)

    wallets = {}
    print("Detecting local wallet keys...")
    try:
        for key in query_keychain_items(cfg.keyring_backend):
            try:
                info = query_keychain_item(key, cfg.keyring_backend)
                wallets[key] = LocalWallet(PrivateKey(info.private_key))
                print(f"Importing wallets['{key}']")
            except Exception:
                print(f"Failed to import local key '{key}'")
        print("Detecting local wallet keys...complete")
    except Exception as ex:
        print(f"No local keychain found: {ex}")

    shell_globals["cfg"] = cfg
    shell_globals["project_path"] = PROJECT_PATH
    shell_globals["profile"] = selected_profile.name
    shell_globals["contracts"] = contracts
    shell_globals["wallets"] = wallets
    for (name, instance) in contract_instances.items():
        shell_globals[name] = instance

    return shell_globals


def run(args: argparse.Namespace):
    cfg = Config.load(PROJECT_PATH)
    profile: Profile = get_profile(cfg, args)
    shell_globals = globals()

    with network_context(profile.network, cfg.project_name, profile.name):
        shell_globals.update(load_shell_globals(cfg, profile))
        embed(shell_globals, vi_mode=False, history_filename=".shell_history")


def add_shell_command(parser):
    shell_cmd = parser.add_parser("shell")
    shell_cmd.add_argument(
        "-p", "--profile", default=None, help="The profile to use"
    )
    shell_cmd.set_defaults(handler=run)



from typing import Dict, Optional, Any, Callable, List
from abc import ABC, abstractmethod
from keyword import iskeyword as is_python_keyword

import grpc
from cosmpy.aerial.client import LedgerClient
from cosmpy.aerial.contract import LedgerContract, _compute_digest
from cosmpy.aerial.tx_helpers import SubmittedTx
from cosmpy.aerial.wallet import Wallet
from cosmpy.crypto.address import Address
from cosmpy.protos.cosmwasm.wasm.v1.query_pb2 import QueryCodeRequest
from jsonschema import ValidationError, validate as validate_schema
from makefun import create_function
from jenesis.contracts import Contract


KEYWORD_PREFIX = '_'
COSMPY_LEDGER_ARGS = [
    'sender',
    'label',
    'gas_limit',
    'store_gas_limit',
    'instantiate_gas_limit',
    'admin_address',
    'funds'
]


class InstantiateArgsError(RuntimeError):
    pass


class ContractObserver(ABC):
    @abstractmethod
    def on_code_id_update(self, code_id: int):
        pass

    @abstractmethod
    def on_contract_address_update(self, address: Address):
        pass


def validate(args: Any, schema: dict):
    try:
        validate_schema(args, schema)
    except ValidationError as ex:
        print("Contract message failed validation. To send to ledger anyway, retry with 'do_validate=False'")
        raise ex


class MonkeyContract(LedgerContract):
    def __init__(
        self,
        contract: Contract,
        client: LedgerClient,
        address: Optional[Address] = None,
        digest: Optional[bytes] = None,
        code_id: Optional[int] = None,
        observer: Optional[ContractObserver] = None,
        init_args: Optional[dict] = None,
    ):
        # pylint: disable=super-init-not-called
        self._contract = contract
        self._path = contract.binary_path
        self._client = client
        self._address = address
        self._observer = observer
        self._init_args = init_args

        self._load_schema(contract.source_path)

        # build the digest
        if contract.binary_path is not None:
            self._digest = _compute_digest(contract.binary_path)
        elif digest is not None:
            self._digest = digest
        else:
            raise RuntimeError('Unable to determine contract digest')

        # look up the code id if this is the first time
        if code_id is None:
            self._code_id = self._find_contract_id_by_digest(self._digest)
        elif code_id <= 0:
            self._code_id = 0
        else:
            self._code_id = self._find_contract_id_by_digest_with_hint(code_id)

        # if code id is not found, store this as code_id = 0 so we don't keep looking for it
        if self._code_id is None:
            self._code_id = 0

        # trigger the observer if necessary
        if self._observer is not None and self._code_id is not None:
            self._observer.on_code_id_update(self._code_id)

    def store(
            self,
            sender: Wallet,
            gas_limit: Optional[int] = None,
            memo: Optional[str] = None,
    ) -> int:
        code_id = super().store(sender, gas_limit=gas_limit, memo=memo)

        # trigger the observer if necessary
        if self._observer is not None:
            self._observer.on_code_id_update(code_id)

        return code_id

    def instantiate(
            self,
            args: Any,
            sender: Wallet,
            label: Optional[str] = None,
            gas_limit: Optional[int] = None,
            admin_address: Optional[Address] = None,
            funds: Optional[str] = None,
            do_validate: Optional[bool] = True,
    ) -> Address:
        # if no args provided, insert init args from configuration
        if args is None:
            if self._init_args is not None:
                args = self._init_args
            else:
                raise InstantiateArgsError(
                    'Please provide instantiation arguments either in "args" or in the jenesis.toml configuration for this contract and profile'
                )
        if do_validate and self._contract.instantiate_schema:
            validate(args, self._contract.instantiate_schema)
        address = super().instantiate(args, sender, label=label, gas_limit=gas_limit,
                                      admin_address=admin_address, funds=funds)

        if self._observer is not None:
            self._observer.on_contract_address_update(address)

        return address

    def _deploy(
        self,
        args: Any,
        sender: Wallet,
        label: Optional[str] = None,
        store_gas_limit: Optional[int] = None,
        instantiate_gas_limit: Optional[int] = None,
        admin_address: Optional[Address] = None,
        funds: Optional[str] = None,
        do_validate: Optional[bool] = True,
    ) -> Address:

        # in the case where the contract is already deployed
        if self._address is not None and self._code_id is not None:
            return self._address

        assert self._address is None

        if self._code_id is None or self._code_id <= 0:
            self.store(sender, gas_limit=store_gas_limit)

        assert self._code_id

        address = self.instantiate(
            args,
            sender,
            label=label,
            gas_limit=instantiate_gas_limit,
            admin_address=admin_address,
            funds=funds,
            do_validate=do_validate,
        )
        return address

    def execute(
        self,
        args: Any,
        sender: Wallet,
        gas_limit: Optional[int] = None,
        funds: Optional[str] = None,
        do_validate: Optional[bool] = True,
    ) -> SubmittedTx:
        if do_validate and self._contract.execute_schema:
            validate(args, self._contract.execute_schema)
        return super().execute(args, sender, gas_limit, funds)

    def query(self, args: Any, do_validate: Optional[bool] = True) -> Any:
        if do_validate and self._contract.query_schema:
            validate(args, self._contract.query_schema)
        return super().query(args)

    def make_queries(self) -> Dict[str, Callable]:

        def make_query(msg: str, msg_args: List[str]):
            def query(self, *args, **kwargs):
                query_arg = {msg: {python_keyword_unwrapper(key): value for (key, value) in kwargs.items()}}
                return self.query(query_arg, *args)

            sig = self._make_function_signature(msg, msg_args, [])
            func = create_function(sig, query)
            return func

        queries = {}
        for (msg, msg_args) in self._contract.query_msgs().items():
            queries[msg] = make_query(msg, msg_args)

        return queries

    def make_executions(self) -> Dict[str, Callable]:

        def make_execution(msg: str, msg_args: List[str]):

            ledger_args = ['sender', 'gas_limit', 'funds']

            def execute(self, **kwargs):
                ledger_exec_args = {}
                contract_func_args = {}
                for (key, value) in kwargs.items():
                    if key in ledger_args:
                        ledger_exec_args[key] = value
                    else:
                        contract_func_args[python_keyword_unwrapper(key)] = value
                execute_arg = {msg: contract_func_args}
                return self.execute(execute_arg, **ledger_exec_args)

            sig = self._make_function_signature(msg, msg_args, ledger_args)
            func = create_function(sig, execute)
            return func

        executions = {}
        for (msg, msg_args) in self._contract.execute_msgs().items():
            executions[msg] = make_execution(msg, msg_args)

        return executions

    def make_deploy(self) -> Dict[str, Callable]:

        def make_deploy(msg: str, msg_args: List[str]):

            ledger_args = [
                'sender', 'label', 'store_gas_limit', 'instantiate_gas_limit', 'admin_address', 'funds'
            ]

            def deploy(self, **kwargs):
                ledger_deploy_args = {}
                contract_init_args = {}
                for (key, value) in kwargs.items():
                    if key in ledger_args:
                        ledger_deploy_args[key] = value
                    else:
                        contract_init_args[python_keyword_unwrapper(key)] = value
                return self._deploy(contract_init_args, **ledger_deploy_args)

            sig = self._make_function_signature(msg, msg_args, ledger_args)
            func = create_function(sig, deploy)
            return func

        deploy = {'deploy': make_deploy('deploy', self._contract.init_args())}

        return deploy

    @staticmethod
    def _make_function_signature(
        name: str,
        msg_args: Dict[str, Any],
        ledger_args: List[str],
    ):
        sig_args = ['self']
        for (arg, description) in msg_args.items():
            if description.get('required'):
                sig_args.append(python_keyword_wrapper(arg))
        for (arg, description) in msg_args.items():
            if not description.get('required'):
                sig_args.append(f"{python_keyword_wrapper(arg)}=None")
        sig_args += [f'{arg}=None' for arg in ledger_args]
        return f'{name}({",".join(sig_args)})'

    def _find_contract_id_by_digest_with_hint(self, code_id_hint: int) -> Optional[int]:

        # try and lookup the specified code id
        try:
            req = QueryCodeRequest(code_id=code_id_hint)
            resp = self._client.wasm.Code(req)

            if resp.code_info.data_hash == self.digest:
                return int(resp.code_info.code_id)
        except (grpc.RpcError, RuntimeError) as ex:
            not_found = False

            # pylint: disable=no-member
            if hasattr(ex, 'details') and 'not found' in ex.details() or 'not found' in str(ex):
                not_found = True

            if not not_found:
                raise

        return self._find_contract_id_by_digest(self._digest)

    def __repr__(self):
        return str(self._address)


def make_contract(
    contract: Contract,
    client: LedgerClient,
    address: Optional[Address] = None,
    digest: Optional[bytes] = None,
    code_id: Optional[int] = None,
    observer: Optional[ContractObserver] = None,
    init_args: Optional[dict] = None,
) -> Any:
    """
    Makes the contract objects for interaction from the shell and scripts.
    The purpose of this factory function is to attach the contract executions
    and queries so that the shell's autocomplete function will work.

    :param contract: The static contract data
    :param client: The client for interacting with the ledger
    :param address: The contract address
    :param digest: The contract digest
    :param code_id: The contract code_id
    :param observer: The contract observer
    :param init_args: The instantiation arguments
    :return: The contract object with queries and executions attached
    """

    monkey_contract = MonkeyContract(
        contract, client, address, digest, code_id, observer, init_args
    )

    # add methods based on schema
    contract_functions = {}
    if contract.schema is not None:
        contract_functions.update(monkey_contract.make_queries())
        contract_functions.update(monkey_contract.make_executions())
        contract_functions.update(monkey_contract.make_deploy())

    JenesisContract = type('JenesisContract', (MonkeyContract,), contract_functions)

    return JenesisContract(
        contract, client, address, digest, code_id, observer, init_args
    )


def _iskeyword(arg: str):
    return is_python_keyword(arg) or arg in COSMPY_LEDGER_ARGS


def python_keyword_wrapper(arg: str) -> str:
    return f'{KEYWORD_PREFIX}{arg}' if _iskeyword(arg) else arg


def python_keyword_unwrapper(arg: str) -> str:
    if KEYWORD_PREFIX in arg:
        if _iskeyword(arg[len(KEYWORD_PREFIX):]):
            return arg[len(KEYWORD_PREFIX):]
    return arg



# pylint: disable=C0103
from ctypes import CDLL, byref, c_char, c_int, c_int32, c_uint32, c_void_p
from ctypes.util import find_library

# Much of the following has been adapted from the keyring library on PyPi:
# https://github.com/jaraco/keyring/blob/3fe6695fabe6bdc7a4f0d72b1fea8f00d62067de/keyring/backends/macOS/api.py
_sec = CDLL(find_library("Security"))
_core = CDLL(find_library("CoreServices"))
_found = CDLL(find_library("Foundation"))

CFDictionaryCreate = _found.CFDictionaryCreate
CFDictionaryCreate.restype = c_void_p
CFDictionaryCreate.argtypes = (
    c_void_p,
    c_void_p,
    c_void_p,
    c_int32,
    c_void_p,
    c_void_p,
)

CFStringCreateWithCString = _found.CFStringCreateWithCString
CFStringCreateWithCString.restype = c_void_p
CFStringCreateWithCString.argtypes = (c_void_p, c_void_p, c_uint32)

CFNumberCreate = _found.CFNumberCreate
CFNumberCreate.restype = c_void_p
CFNumberCreate.argtypes = (c_void_p, c_uint32, c_void_p)

SecItemAdd = _sec.SecItemAdd
SecItemAdd.restype = c_int32
SecItemAdd.argtypes = (c_void_p, c_void_p)

SecItemCopyMatching = _sec.SecItemCopyMatching
SecItemCopyMatching.restype = c_int32
SecItemCopyMatching.argtypes = (c_void_p, c_void_p)

SecItemDelete = _sec.SecItemDelete
SecItemDelete.restype = c_int32
SecItemDelete.argtypes = (c_void_p,)

CFDataGetBytePtr = _found.CFDataGetBytePtr
CFDataGetBytePtr.restype = c_void_p
CFDataGetBytePtr.argtypes = (c_void_p,)

CFDataGetLength = _found.CFDataGetLength
CFDataGetLength.restype = c_int32
CFDataGetLength.argtypes = (c_void_p,)

CFArrayGetCount = _found.CFArrayGetCount
CFArrayGetCount.restype = c_int
CFArrayGetCount.argtypes = (c_void_p,)

CFArrayGetValueAtIndex = _found.CFArrayGetValueAtIndex
CFArrayGetValueAtIndex.restype = c_void_p
CFArrayGetValueAtIndex.argtypes = (c_void_p, c_int)

CFDictionaryGetCount = _found.CFDictionaryGetCount
CFDictionaryGetCount.restype = c_int
CFDictionaryGetCount.argtypes = (c_void_p,)

CFDictionaryGetKeysAndValues = _found.CFDictionaryGetKeysAndValues
CFDictionaryGetKeysAndValues.restype = c_void_p
CFDictionaryGetKeysAndValues.argtypes = (c_void_p, c_void_p, c_void_p)

CFDictionaryGetValue = _found.CFDictionaryGetValue
CFDictionaryGetValue.restype = c_void_p
CFDictionaryGetValue.argtypes = (c_void_p, c_void_p)

CFShow = _found.CFShow
CFShow.restype = c_void_p
CFShow.argtypes = (c_void_p,)

CFStringGetLength = _found.CFStringGetLength
CFStringGetLength.restype = c_int
CFStringGetLength.argtypes = (c_void_p,)

CFStringGetCStringPtr = _found.CFStringGetCStringPtr
CFStringGetCStringPtr.restype = c_int
CFStringGetCStringPtr.argtypes = (c_void_p, c_uint32)

CFStringGetCString = _found.CFStringGetCString
CFStringGetCString.restype = c_int
CFStringGetCString.argtypes = (c_void_p, c_void_p, c_int, c_uint32)

# String encodings
kCFStringEncodingASCII = 0x0600
kCFStringEncodingUTF8 = 0x08000100


def k_(s):
    return c_void_p.in_dll(_sec, s)


def create_cfbool(b: bool):
    return CFNumberCreate(None, 0x9, byref(c_int32(1 if b else 0)))  # int32


def create_cfstr(s: str):
    return CFStringCreateWithCString(None, s.encode("utf8"), kCFStringEncodingUTF8)


def get_cfstr_value(item: c_void_p) -> str:
    # determine the string length (and double it because CFStringGetCString requires extra space)
    buffer_len = CFStringGetLength(item) * 2

    # need to create a large string buffer here because otherwise this process can fail
    buffer = (c_char * buffer_len)()
    status = CFStringGetCString(item, byref(buffer), buffer_len, kCFStringEncodingUTF8)
    if status != 1:
        raise RuntimeError("Unable to get CFString value")

    return buffer.value.decode()



import argparse
import os
from typing import Optional

from jenesis.config import Config
from jenesis.network import is_local_node_running, run_local_node, LedgerNodeDockerContainer


def get_profile(profile_name: Optional[str]):
    project_path = os.getcwd()

    # check that we are actually running the command from the project root
    if not os.path.exists(os.path.join(project_path, "jenesis.toml")):
        print("Please run command from project root")
        return None

    cfg = Config.load(project_path)
    return cfg, cfg.get_profile(profile_name)


def run_start(args: argparse.Namespace):
    cfg, profile = get_profile(args.profile)

    if profile is None:
        return 1

    if not profile.network.is_local:
        print("This profile is configured for a remote network.")
        return 1
    if run_local_node(profile.network, cfg.project_name, profile.name):
        return 0
    return 1


def run_stop(args: argparse.Namespace):
    cfg, profile = get_profile(args.profile)

    if profile is None:
        return 1

    if not profile.network.is_local:
        print("This profile is configured for a remote network.")
        return 1
    local_node = LedgerNodeDockerContainer(profile.network, cfg.project_name, profile.name)
    if local_node.container is not None:
        if local_node.is_ready():
            print("Shutting down local_node...")
            local_node.container.stop()
            print("Shutting down local_node...complete")
            return 0
        if is_local_node_running():
            print("The currently running local node does not match this project and profile. To stop this node, please stop it from the appropriate project and profile or by using Docker.")
    else:
        print("Did not find any local node for this profile")
    return 1


def get_logs(local_node: LedgerNodeDockerContainer):
    for line in local_node.container.logs(stream=True):
        yield line.strip().decode()


def run_logs(args: argparse.Namespace):
    cfg, profile = get_profile(args.profile)

    if profile is None:
        return 1

    if not profile.network.is_local:
        print("This profile is configured for a remote network.")
        return 1

    local_node = LedgerNodeDockerContainer(profile.network, cfg.project_name, profile.name)
    if local_node.container is not None:
        for line in get_logs(local_node):
            print(line)
    return 1


def add_network_command(parser):
    network_parser = parser.add_parser("network", help="Start or stop a local node")
    subparsers = network_parser.add_subparsers()

    netstart_cmd = subparsers.add_parser("start", help="Start local node")
    netstop_cmd = subparsers.add_parser("stop", help="Stop local node")
    netlogs_cmd = subparsers.add_parser("logs", help="Show local node logs")
    for cmd in [netstart_cmd, netstop_cmd, netlogs_cmd]:
        cmd.add_argument(
            "-p", "--profile", default=None, help="The profile associated with the network"
        )
    netstart_cmd.set_defaults(handler=run_start)
    netstop_cmd.set_defaults(handler=run_stop)
    netlogs_cmd.set_defaults(handler=run_logs)



from ctypes import byref, c_void_p, string_at
from io import BytesIO
from typing import List, Union

from jenesis.keyring.amino.codec import (
    LocalInfo,
    OfflineInfo,
    unmarshal_binary_length_prefixed,
)
from jenesis.keyring.macos.abi import (
    CFArrayGetCount,
    CFArrayGetValueAtIndex,
    CFDataGetBytePtr,
    CFDataGetLength,
    CFDictionaryCreate,
    CFDictionaryGetValue,
    CFShow,
    CFStringGetCString,
    CFStringGetLength,
    SecItemCopyMatching,
    _found,
    create_cfbool,
    create_cfstr,
    get_cfstr_value,
    k_,
)
from jenesis.keyring.errors import KeychainError


class MacOsKeychainError(KeychainError):
    pass


def _create_query(**kwargs):
    return CFDictionaryCreate(
        None,
        (c_void_p * len(kwargs))(*[k_(k) for k in kwargs]),
        (c_void_p * len(kwargs))(
            *[create_cfstr(v) if isinstance(v, str) else v for v in kwargs.values()]
        ),
        len(kwargs),
        _found.kCFTypeDictionaryKeyCallBacks,
        _found.kCFTypeDictionaryValueCallBacks,
    )


def query_keychain_item(name: str) -> Union[LocalInfo, OfflineInfo]:
    query = _create_query(
        kSecClass=k_("kSecClassGenericPassword"),
        kSecMatchLimit=k_("kSecMatchLimitOne"),
        kSecAttrService="fetch",
        kSecAttrAccount=f"{name}.info",
        kSecReturnData=create_cfbool(True),
    )

    data = c_void_p()
    status = SecItemCopyMatching(query, byref(data))

    if status != 0:
        raise MacOsKeychainError("Unable to query keyring items")

    buffer = BytesIO(string_at(CFDataGetBytePtr(data), CFDataGetLength(data)))
    info = unmarshal_binary_length_prefixed(buffer)
    if info is None:
        raise MacOsKeychainError("Unable to parse stored entry")

    return info


def query_keychain_items() -> List[str]:
    query = _create_query(
        kSecClass=k_("kSecClassGenericPassword"),
        kSecMatchLimit=k_("kSecMatchLimitAll"),
        kSecReturnAttributes=create_cfbool(True),
        kSecAttrService="fetch",
    )

    data = c_void_p()
    status = SecItemCopyMatching(query, byref(data))

    if status != 0:
        raise MacOsKeychainError("Unable to query keyring items")

    key_names = []

    # iterate over all the items
    for idx in range(CFArrayGetCount(data)):

        # lookup the CFDictionary at the given index
        val = CFArrayGetValueAtIndex(data, idx)

        # attempt to find the account ('acct') field for each item
        item = CFDictionaryGetValue(val, create_cfstr("acct"))

        # skip all items that don't have an acct field
        if item == 0:
            continue

        account = get_cfstr_value(item)
        if account.endswith(".info"):
            key_names.append(account[:-5])

    return list(set(key_names))



from abc import abstractmethod

from docker.models.containers import Container
from docker.errors import DockerException
from jenesis.tasks import Task, TaskStatus

class ContainerTask(Task):

    def __init__(self):
        self._build_image = None
        self._container = None
        self._status = TaskStatus.IDLE
        self._status_text = ''
        self._in_progress_text = 'Building...'
        self._logs = ''

    @property
    def logs_text(self) -> str:
        return self._logs

    @property
    def status_text(self) -> str:
        return self._status_text

    @property
    def status(self) -> TaskStatus:
        return self._status

    def poll(self):
        # no further processing required if it has either completed or failed
        if self.is_done:
            return

        # if the output is not out of date then the container task is unnecessary
        if self._container is None and not self._is_out_of_date():
            self._status = TaskStatus.COMPLETE
            self._status_text = ''
            return

        # if we get this far we either need to schedule the container or monitor progress
        if self._container is None:
            try:
                self._container = self._schedule_container()
                self._status = TaskStatus.IN_PROGRESS
                self._status_text = self._in_progress_text
            except DockerException:
                print("Error: looks like your docker setup isn't right, please visit https://jenesis.fetch.ai/ for more information")
                self._container = None
                self._status = TaskStatus.FAILED
                self._status_text = ''

            # exit if we do not have docker or some such installed
            if self._container is None:
                return

        assert self._container is not None

        # check on the progress of the container
        self._container.reload()

        if self._show_logs():
            log_text = self._container.logs().decode("utf-8")
            self._logs = log_text

        if self._container.status == 'exited':
            exit_code = int(self._container.attrs['State']['ExitCode'])
            if exit_code == 0:
                self._status = TaskStatus.COMPLETE
                self._status_text = ''
            else:
                self._status = TaskStatus.FAILED
                self._status_text = ''
                log_text = self._container.logs().decode("utf-8")
                self._logs = log_text
            self._container.remove()

    def teardown(self):
        if self._container and self._status == TaskStatus.IN_PROGRESS:
            print(f'Stopping build container for {self.name}...')
            self._container.kill()
            print(f'Removing build container for {self.name}...')
            self._container.remove()

    @abstractmethod
    def _is_out_of_date(self) -> bool:
        pass

    @abstractmethod
    def _schedule_container(self) -> Container:
        pass

    @abstractmethod
    def _show_logs(self):
        pass



import argparse
import os
from jenesis.config import Config


def run(args: argparse.Namespace):
    project_path = os.getcwd()

    # check if the project has been already initialized
    if os.path.exists(os.path.join(project_path, args.project)):
        print("Project already exists")
        return

    Config.create_project(args.project, args.profile, args.network)

def add_new_command(parser):
    new_cmd = parser.add_parser("new")
    new_cmd.add_argument("project", help="Project name")
    new_cmd.add_argument(
        "-p", "--profile", default="testing", help="The profile to create"
    )
    new_cmd.add_argument(
        "-n", "--network", default="fetchai-testnet", help="Network to use"
    )
    new_cmd.set_defaults(handler=run)



import os.path
from dataclasses import dataclass
from typing import Optional

from cosmpy.aerial.contract import _compute_digest


@dataclass
class Contract:
    name: str
    source_path: str
    binary_path: str
    cargo_root: str
    schema: dict

    def __post_init__(self):
        self.instantiate_schema = {}
        self.query_schema = {}
        self.execute_schema = {}

        self.update_schema()


    def update_schema(self):
        # check for workspace-style schemas
        if self.name in self.schema:
            contract_schema = self.schema[self.name]
        else:
            contract_schema = self.schema

        for (msg_type, schema) in contract_schema.items():
            if 'instantiate' in msg_type:
                self.instantiate_schema = schema
            elif 'query' in msg_type:
                self.query_schema = schema
            elif 'execute' in msg_type:
                self.execute_schema = schema

    def digest(self) -> Optional[str]:
        if not os.path.isfile(self.binary_path):
            return None

        return _compute_digest(self.binary_path).hex()

    def execute_msgs(self) -> dict:
        return _extract_msgs(self.execute_schema)

    def query_msgs(self) -> dict:
        return _extract_msgs(self.query_schema)

    def init_args(self) -> dict:
        return _get_msg_args(self.instantiate_schema)

    def __repr__(self):
        return self.name


def _extract_msgs(schema: dict) -> dict:
    msgs = {}
    if 'oneOf' in schema:
        schemas = schema['oneOf']
    elif 'anyOf' in schema:
        schemas = schema['anyOf']
    else:
        return msgs
    for msg_schema in schemas:
        msg = msg_schema['required'][0]
        args = _get_msg_args(msg_schema['properties'][msg])
        msgs[msg] = args
    return msgs

def _get_msg_args(msg_schema):
    msg_args = {}
    args = msg_schema.get('properties', {}).keys()
    required = msg_schema.get('required', [])
    for arg in args:
        description = msg_schema['properties'][arg]
        msg_args[arg] = {
            'required': arg in required,
            'type': description.get('type') or description.get('$ref')
        }
    return msg_args



import multiprocessing
from typing import Optional

from docker.client import from_env
from docker.models.images import ImageCollection
from docker.errors import DockerException, ImageNotFound
from jenesis.tasks import Task, TaskStatus
from jenesis.tasks.monitor import run_tasks


multiprocessing.set_start_method('fork')

image_collection = ImageCollection(from_env())


class ImagePullTask(Task):

    def __init__(self, image: str):
        self._image = image
        self._process: Optional[multiprocessing.Process] = None
        self._status = TaskStatus.IDLE
        self._status_text = ''

    @property
    def name(self) -> str:
        return self._image

    @property
    def logs_text(self) -> str:
        return ''

    @property
    def status_text(self) -> str:
        return self._status_text

    @property
    def status(self) -> TaskStatus:
        return self._status

    def poll(self):
        # no further processing required if it has either completed or failed
        if self.is_done:
            return

        # if the image already exists then the task is completed or unnecessary
        if image_exists(self._image):
            self._status = TaskStatus.COMPLETE
            self._status_text = ''
            return

        # if we get this far we either need to start the image pull or monitor progress
        if self._process is None:
            try:
                self._process = multiprocessing.Process(
                    target=image_collection.pull,
                    name=f'ImagePull-{self._image}',
                    args=[self._image],
                )
                self._process.start()
                self._status = TaskStatus.IN_PROGRESS
                self._status_text = 'Pulling image...'
            except DockerException:
                print("Error: looks like your docker setup isn't right, please visit https://jenesis.fetch.ai/ for more information")
                self._status = TaskStatus.FAILED
                self._status_text = ''
        else:
            if not self._process.is_alive():
                self._status = TaskStatus.FAILED
                self._status_text = ''

    def teardown(self):
        if self._process is not None:
            if self._process.is_alive():
                self._process.terminate()


def image_exists(image):
    try:
        image_collection.get(image)
        return True
    except ImageNotFound:
        return False


def pull_image(
    image: str,
):
    """
    Will attempt to pull the appropriate docker image

    :return:
    """
    # create all the tasks to be done
    tasks = [ImagePullTask(image)]

    # run the tasks
    run_tasks(tasks)



import hashlib
import io
from dataclasses import dataclass
from typing import IO, Generator, Optional, Tuple, Union

LOCAL_INFO_PREFIX = (
    b"\x0d\xad\x15\x3d"  # = build_prefix_from_string('crypto/keys/localInfo')
)
OFFLINE_INFO_PREFIX = (
    b"\x1f\x4d\x3e\xd6"  # = build_prefix_from_string('crypto/keys/offlineInfo')
)
PUBLIC_KEY_SECP256K1_PREFIX = (
    b"\xeb\x5a\xe9\x87"  # = build_prefix_from_string('tendermint/PubKeySecp256k1')
)
PRIVATE_KEY_SECP256K1_PREFIX = (
    b"\xe1\xb0\xf7\x9b"  # = build_prefix_from_string('tendermint/PrivKeySecp256k1')
)
PUBLIC_KEY_SECP256K1_LENGTH = 33
PRIVATE_KEY_SECP256K1_LENGTH = 32


@dataclass
class OfflineInfo:
    name: str
    public_key: bytes
    algorithm: str


@dataclass
class LocalInfo:
    name: str
    public_key: bytes
    private_key: bytes
    algorithm: str


class CodecParseError(RuntimeError):
    pass


def compute_prefix(data: bytes) -> bytes:
    assert len(data) == 32  # should match a sha256 digest

    if data[0] == 0:
        skip_length = 4
    else:
        skip_length = 3

    if data[skip_length] == 0:
        skip_length += 1

    return data[skip_length : skip_length + 4]


def build_prefix_from_string(name: str) -> bytes:
    hasher = hashlib.sha256()
    hasher.update(name.encode())
    return compute_prefix(hasher.digest())


def parse_uvarint(data: IO[bytes]) -> int:
    value = 0
    while True:
        data_bytes = data.read(1)
        if len(data_bytes) == 0:
            raise CodecParseError("Reached the end of the binary stream")

        # add the bits to the value
        value = (value << 7) | (data_bytes[0] & 0x7F)

        # determine if there is a continuation mark on this byte
        if data_bytes[0] & 0x80 == 0:
            break

    return value


def extract_bytes(data: IO[bytes], length: int) -> bytes:
    value = data.read(length)
    if len(value) == 0:
        raise CodecParseError("reached the EOF prematurely during processing")
    if len(value) != length:
        raise CodecParseError(
            f"unable to read expected number of bytes (wanted: {length} received: {len(value)}"
        )
    return value


def extract_byte(data: IO[bytes]) -> int:
    value = extract_bytes(data, 1)
    return value[0]


def unmarshal_fields(data: IO[bytes]) -> Generator[Tuple[int, bytes], None, None]:
    while True:
        prefix = data.read(1)
        if prefix == b"":
            break

        # extract the field and type
        field_index = (prefix[0] >> 3) & 0x1F
        field_type = prefix[0] & 0x7
        if field_type != 2:
            raise CodecParseError(f"unexpected field type {field_type}")

        # extract the length
        length = parse_uvarint(data)
        field_buffer = extract_bytes(data, length)

        yield field_index, field_buffer


def unmarshal_public_key(data: IO[bytes]) -> bytes:
    prefix = extract_bytes(data, 4)
    if prefix != PUBLIC_KEY_SECP256K1_PREFIX:
        raise CodecParseError(f"invalid public key prefix 0x{prefix.hex()}")

    length = parse_uvarint(data)
    if length != PUBLIC_KEY_SECP256K1_LENGTH:
        raise CodecParseError(f"invalid public key length {length}")

    return extract_bytes(data, length)


def unmarshal_private_key(data: IO[bytes]) -> bytes:
    prefix = extract_bytes(data, 4)
    if prefix != PRIVATE_KEY_SECP256K1_PREFIX:
        raise CodecParseError(f"invalid private key prefix 0x{prefix.hex()}")

    length = parse_uvarint(data)
    if length != PRIVATE_KEY_SECP256K1_LENGTH:
        raise CodecParseError(f"invalid private key length {length}")

    return extract_bytes(data, length)


def unmarshal_local_info(data: IO[bytes]) -> LocalInfo:
    parameters = {}

    for idx, field in unmarshal_fields(data):
        if idx == 1:
            parameters["name"] = field.decode()
        elif idx == 2:
            parameters["public_key"] = unmarshal_public_key(io.BytesIO(field))
        elif idx == 3:
            parameters["private_key"] = unmarshal_private_key(io.BytesIO(field))
        elif idx == 4:
            parameters["algorithm"] = field.decode()
        else:
            raise CodecParseError(f"unexpected field index {idx} for LocalInfo")

    return LocalInfo(**parameters)


def unmarshal_offline_info(data: IO[bytes]) -> OfflineInfo:
    parameters = {}

    for idx, field in unmarshal_fields(data):
        if idx == 1:
            parameters["name"] = field.decode()
        elif idx == 2:
            parameters["public_key"] = unmarshal_public_key(io.BytesIO(field))
        elif idx == 3:
            parameters["algorithm"] = field.decode()
        else:
            raise CodecParseError(f"unexpected field index {idx} for LocalInfo")

    return OfflineInfo(**parameters)


def unmarshal_binary_bare(data: IO[bytes]) -> Optional[Union[LocalInfo, OfflineInfo]]:
    prefix = extract_bytes(data, 4)
    if prefix == LOCAL_INFO_PREFIX:
        return unmarshal_local_info(data)
    if prefix == OFFLINE_INFO_PREFIX:
        return unmarshal_offline_info(data)

    return None


def unmarshal_binary_length_prefixed(
    data: IO[bytes],
) -> Optional[Union[LocalInfo, OfflineInfo]]:
    length = parse_uvarint(data)
    raw_data = extract_bytes(data, length)

    # create a new buffer to be processed (we assume that it will be consumed)
    buffer = io.BytesIO(raw_data)
    return unmarshal_binary_bare(buffer)



import argparse
import os

from cosmpy.aerial.client import LedgerClient
from jenesis.config import Config
from jenesis.contracts.detect import parse_contract
from jenesis.contracts.monkey import MonkeyContract
from jenesis.network import Network
from jenesis.network import network_context
import toml

def run(args: argparse.Namespace):

    project_path = os.getcwd()

    # check that we are actually running the command from the project root
    if not os.path.exists(os.path.join(project_path, "jenesis.toml")):
        # pylint: disable=all
        print("Please run command from project root")
        return 1

    cfg = Config.load(project_path)

    if args.profile is not None and args.profile not in cfg.profiles:
        print(f'Invalid profile name. Expected one of {",".join(cfg.profiles.keys())}')
        return 1

    profile_name = args.profile or cfg.get_default_profile()

    selected_profile = cfg.profiles[profile_name]

    if args.deployment not in selected_profile.deployments:
        print('Deployment not found in project')
        return 1

    deployment = selected_profile.deployments[args.deployment]

    client = LedgerClient(selected_profile.network)
    contract_to_attach = parse_contract(project_path, deployment.contract)

    data = toml.load("jenesis.toml")
    network = Network(**data["profile"][profile_name]["network"])

    with network_context(network, cfg.project_name, profile_name):
        contract = MonkeyContract(contract_to_attach, client, args.address)
        code_id = contract.code_id
        digest = contract.digest.hex()

    Config.update_project(project_path, profile_name, network.name, contract_to_attach, deployment.name)

    cfg.update_deployment(
        selected_profile.name, deployment.name, digest, code_id, args.address
    )
    cfg.save(project_path)
    return 0


def add_attach_command(parser):
    attach_cmd = parser.add_parser("attach")
    attach_cmd.add_argument("deployment", help="Deployment name")
    attach_cmd.add_argument("address", help="Contract address")
    attach_cmd.add_argument(
        "-p",
        "--profile",
        default=None,
        help="Profile where deployment will be attached to",
    )
    attach_cmd.set_defaults(handler=run)


from contextlib import contextmanager
import os
import time
import stat
from typing import Optional, List

from docker import from_env
from docker.errors import DockerException
from docker.models.containers import Container

from cosmpy.aerial.config import NetworkConfig
from cosmpy.aerial.client import LedgerClient

DEFAULT_DOCKER_IMAGE_TAG = "fetchai/fetchd:0.10.5"
DEFAULT_VALIDATOR_KEY_NAME = "validator"
DEFAULT_MNEMONIC = "gap bomb bulk border original scare assault pelican resemble found laptop skin gesture height inflict clinic reject giggle hurdle bubble soldier hurt moon hint"
DEFAULT_PASSWORD = "12345678"
DEFAULT_MONIKER = "test-node"
DEFAULT_CHAIN_ID = "testing"
DEFAULT_GENESIS_ACCOUNT = "fetch1gns5lphdk5ew5lnre7ulzv8s8k9dr9eyqvgj0w"
DEFAULT_DENOMINATION = "atestfet"
DEFAULT_CLI_BINARY = "fetchd"
DEFAULT_TIMEOUT_COMMIT = "5s"


LOCALNODE_CONFIG_DIR = os.path.join(os.getcwd(), ".localnode")


class Network(NetworkConfig):

    def __init__(
        self,
        name: str = "",
        chain_id: str = "",
        fee_minimum_gas_price: int = 0,
        fee_denomination: str = "",
        staking_denomination: str = "",
        url: str = "",
        faucet_url: Optional[str] = None,
        is_local: Optional[bool] = False,
        keep_running: Optional[bool] = False,
        cli_binary: Optional[str] = None,
        validator_key_name: Optional[str] = None,
        mnemonic: Optional[str] = None,
        password: Optional[str] = None,
        moniker: Optional[str] = None,
        genesis_accounts: Optional[List[str]] = None,
        timeout_commit: Optional[str] = None,
        debug_trace: bool = True,
    ):
        super().__init__(
            chain_id,
            fee_minimum_gas_price,
            fee_denomination,
            staking_denomination,
            url,
            faucet_url=faucet_url,
        )
        self.name = name
        self.is_local = is_local
        if is_local:
            self.keep_running = keep_running
            self.cli_binary = cli_binary or DEFAULT_CLI_BINARY
            self.validator_key_name = validator_key_name or DEFAULT_VALIDATOR_KEY_NAME
            self.mnemonic = mnemonic or DEFAULT_MNEMONIC
            self.password = password or DEFAULT_PASSWORD
            self.moniker = moniker or DEFAULT_MONIKER
            self.genesis_accounts = genesis_accounts or [DEFAULT_GENESIS_ACCOUNT]
            self.timeout_commit = timeout_commit or DEFAULT_TIMEOUT_COMMIT
            self.debug_trace = debug_trace

    @classmethod
    def from_cosmpy_config(cls, name: str, net_config: NetworkConfig, is_local: bool = False):
        return Network(
            name,
            net_config.chain_id,
            net_config.fee_minimum_gas_price,
            net_config.fee_denomination,
            net_config.staking_denomination,
            net_config.url,
            faucet_url=net_config.faucet_url,
            is_local=is_local,
        )


class LedgerNodeDockerContainer:

    PORTS = {9090: 9090, 1317: 1317, 26657: 26657}

    def __init__(
        self,
        network: Network,
        project_name: str,
        profile_name: str,
        tag: str = DEFAULT_DOCKER_IMAGE_TAG,
    ):
        """
        Initialize the Fetch ledger Docker image.

        :param client: the Docker client.
        :param addr: the address.
        :param port: the port.
        :param config: optional configuration to command line.
        """
        self._client = from_env()
        self._image_tag = tag
        self.name = f"{network.name}-{project_name}-{profile_name}"
        self.network: Network = network

        self.container: Optional[Container] = None
        try:
            self.container: Container = self._client.containers.get(self.name)
        except Exception:
            pass

    @property
    def tag(self) -> str:
        """Get the image tag."""
        return self._image_tag

    def _make_entrypoint_script(self) -> List[str]:
        """Make a temporary entrypoint file to setup and run the test ledger node"""
        trace_flag = '--trace' if self.network.debug_trace else ''
        entrypoint_lines = [
            "#!/usr/bin/env bash",
            'if [ ! -f /root/.fetchd/config/genesis.json ]; then',
            # variables
            f'export VALIDATOR_KEY_NAME={self.network.validator_key_name}',
            f'export VALIDATOR_MNEMONIC="{self.network.mnemonic}"',
            f'export PASSWORD="{self.network.password}"',
            f'export CHAIN_ID={self.network.chain_id}',
            f'export MONIKER={self.network.moniker}',
            f'export DENOM={self.network.fee_denomination}',
            # Add key
            f'( echo "$VALIDATOR_MNEMONIC"; echo "$PASSWORD"; echo "$PASSWORD"; ) |{self.network.cli_binary} keys add $VALIDATOR_KEY_NAME --recover',
            # Configure node
            f"{self.network.cli_binary} init --chain-id=$CHAIN_ID $MONIKER",
        ]
        for acc in self.network.genesis_accounts:
            entrypoint_lines.append(
                f'echo "$PASSWORD" |{self.network.cli_binary} add-genesis-account {acc} 100000000000000000000000$DENOM',
            )
        entrypoint_lines.extend([
            f'echo "$PASSWORD" |{self.network.cli_binary} add-genesis-account $({self.network.cli_binary} keys show $VALIDATOR_KEY_NAME -a) 100000000000000000000000$DENOM',
            f'echo "$PASSWORD" |{self.network.cli_binary} gentx $VALIDATOR_KEY_NAME 10000000000000000000000$DENOM --chain-id $CHAIN_ID',
            f"{self.network.cli_binary} collect-gentxs",
            f'sed -i "s/stake/{self.network.fee_denomination}/" ~/.{self.network.cli_binary}/config/genesis.json',
            f'''sed -i 's/timeout_commit = "5s"/timeout_commit = "{self.network.timeout_commit}"/' ~/.{self.network.cli_binary}/config/config.toml''',
            # Enable rest-api
            f'sed -i "s/enable = false/enable = true/" ~/.{self.network.cli_binary}/config/app.toml',
            f'sed -i "s/swagger = false/swagger = true/" ~/.{self.network.cli_binary}/config/app.toml',
            'fi',
            f"{self.network.cli_binary} start --rpc.laddr tcp://0.0.0.0:26657 {trace_flag}",
        ])
        return [line + "\n" for line in entrypoint_lines]

    def update_entrypoint_file(self) -> bool:

        if not os.path.isdir(LOCALNODE_CONFIG_DIR):
            os.mkdir(LOCALNODE_CONFIG_DIR)

        previous_entrypoint_script = []
        entrypoint_file = os.path.join(LOCALNODE_CONFIG_DIR, "run-node.sh")
        if os.path.isfile(entrypoint_file):
            with open(entrypoint_file, encoding="utf-8") as file:
                previous_entrypoint_script = file.readlines()

        entrypoint_script = self._make_entrypoint_script()
        entrypoint_outdated = entrypoint_script != previous_entrypoint_script

        if entrypoint_outdated:
            with open(entrypoint_file, "w", encoding="utf-8") as file:
                file.writelines(line for line in entrypoint_script)
            os.chmod(entrypoint_file, stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)

        return entrypoint_outdated

    def run(self):

        entrypoint_outdated = self.update_entrypoint_file()

        if self.container:
            if entrypoint_outdated:
                self.container.remove()
                self.container = None
            else:
                self.container.start()
                return

        # if the container doesn't exist, create a new container and run it
        mount_path = "/mnt"
        volumes = {LOCALNODE_CONFIG_DIR: {"bind": mount_path, "mode": "rw"}}
        entrypoint = os.path.join(mount_path, "run-node.sh")
        self.container = self._client.containers.run(
            self.tag,
            detach=True,
            volumes=volumes,
            entrypoint=str(entrypoint),
            ports=self.PORTS,
            name=self.name,
        )

    def is_ready(self) -> bool:
        try:
            self.container.reload()
            assert self.container.status == "running"
            assert is_local_node_running()
            return True
        except Exception:
            return False

    def wait_until_ready(self, max_attempts: int = 15, sleep_rate: float = 1.0) -> bool:
        for _ in range(max_attempts):
            if self.is_ready():
                return True
            time.sleep(sleep_rate)
        return False


def run_local_node(network: Network, project_name: str, profile_name: str) -> Optional[LedgerNodeDockerContainer]:
    try:
        local_node = LedgerNodeDockerContainer(network, project_name, profile_name)
        if is_local_node_running():
            if local_node.is_ready():
                print(f"Detected local node running: {local_node.name}")
            else:
                print("Warning: Detected local node running that may not have been configured for this project and profile. If this is not desired, please stop the currently running node and rerun the command.")
        else:
            print("Starting local node...")
            local_node.run()
            if not local_node.wait_until_ready():
                raise RuntimeError('Failed to start local node.')
            print("Stating local node...complete")
            return local_node
    except DockerException as ex:
        print(f"Failed to start local node: {ex}")
    return None


@contextmanager
def network_context(network: Network, project_name: str, profile_name: str):
    if network.is_local:
        local_node = run_local_node(network, project_name, profile_name)
    else:
        local_node = None
    try:
        yield local_node
    finally:
        if local_node:
            if network.keep_running:
                print(f"Local node still running in background: {local_node.name}")
            else:
                print("Shutting down local_node...")
                local_node.container.stop()
                print("Shutting down local_node...complete")


def is_local_node_running():
    try:
        net_config = fetchai_localnode_config()
        client = LedgerClient(net_config)
        validators = client.query_validators()
        assert len(validators) > 0
        return True
    except Exception:
        return False


def fetchai_testnet_config() -> Network:
    return Network.from_cosmpy_config("fetchai-testnet", NetworkConfig.fetchai_stable_testnet())


def fetchai_mainnet_config() -> Network:
    return Network.from_cosmpy_config("fetchai-mainnet", NetworkConfig.fetchai_mainnet())


def fetchai_localnode_config() -> Network:
    return Network.from_cosmpy_config(
        "fetchai-localnode",
        NetworkConfig(
            chain_id="localnode",
            url="grpc+http://127.0.0.1:9090/",
            fee_minimum_gas_price=5000000000,
            fee_denomination="atestfet",
            staking_denomination="atestfet",
        ),
        is_local=True,
    )



import os
from typing import List, Optional

from docker import from_env
from docker.models.containers import Container
from docker.types import Mount

from jenesis.contracts import Contract
from jenesis.tasks.container import ContainerTask
from jenesis.tasks.monitor import run_tasks
from jenesis.tasks.utils import chunks, get_last_modified_timestamp


CONTRACT_BUILD_IMAGE = "cosmwasm/rust-optimizer:0.12.9"
WORKSPACE_BUILD_IMAGE = "cosmwasm/workspace-optimizer:0.12.9"
DEFAULT_BUILD_STEPS = [
    "RUSTFLAGS='-C link-arg=-s' cargo build --release --lib --target wasm32-unknown-unknown",
    "mkdir -p artifacts",
    "mv target/wasm32-unknown-unknown/release/*.wasm artifacts/",
]


class ContractBuildTask(ContainerTask):

    def __init__(self, contract: Contract, optimize: bool, rebuild: bool, log: bool):
        super().__init__()
        self.contract = contract
        self._optimize = optimize
        self._rebuild = rebuild
        self._log = log
        self._build_image = CONTRACT_BUILD_IMAGE
        self._build_steps = DEFAULT_BUILD_STEPS
        self._working_dir = '/code'
        self._in_progress_text = 'Building...'

    @property
    def name(self) -> str:
        return self.contract.name

    def _is_out_of_date(self) -> bool:
        #  pylint: disable=duplicate-code
        if self._rebuild:
            return True

        # determine the timestamp of the compiled contract
        if os.path.isfile(self.contract.binary_path):
            compiled_contract_timestamp = os.path.getmtime(self.contract.binary_path)
        else:
            compiled_contract_timestamp = 0

        # determine the timestamp of the contract source
        src_path = os.path.join(self.contract.source_path, 'src')
        contract_source_timestamp = get_last_modified_timestamp([src_path], 'rs')

        return contract_source_timestamp > compiled_contract_timestamp

    def _schedule_container(self) -> Container:
        mounts = [
            Mount('/code/target', f'contract_{self.contract.name}_cache'),
            Mount('/usr/local/cargo/registry', 'registry_cache'),
            Mount('/code', os.path.abspath(self.contract.cargo_root), type='bind'),
        ]

        # get the docker client
        client = from_env()

        # start the container
        entrypoint = None if self._optimize else "/bin/sh"
        args = None if self._optimize else ["-c", " && ".join(self._build_steps)]
        return client.containers.run(
            self._build_image,
            args,
            mounts=mounts,
            working_dir=self._working_dir,
            entrypoint=entrypoint,
            detach=True,
        )

    def _show_logs(self):
        return self._log


def build_contracts(
    contracts: List[Contract],
    batch_size: Optional[int] = None,
    optimize: Optional[bool] = False,
    rebuild: Optional[bool] = False,
    log: Optional[bool] = False,
):
    """
    Will attempt to build all the specified contracts (provided they are out of date)

    :param contracts: The list of contracts to build
    :param batch_size: The max number of builds to do in parallel. If None then will attempt to all in parallel
    :param optimize: Whether to perform an optimized build
    :param rebuild: Whether to force a rebuild of the contracts
    :return:
    """

    # create all the tasks to be done
    tasks = list(
        map(
            ContractBuildTask,
            contracts,
            [optimize] * len(contracts),
            [rebuild] * len(contracts),
            [log] * len(contracts),
        )
    )

    # run the tasks (in batches if configured)
    for batch in chunks(tasks, batch_size=batch_size):
        run_tasks(batch)


class WorkspaceBuildTask(ContainerTask):

    def __init__(self, path: str, contracts: List[Contract], optimize: bool, rebuild: bool, log: bool):
        super().__init__()
        self._path = path
        self._contracts = contracts
        self._optimize = optimize
        self._rebuild = rebuild
        self._log = log
        self._build_image = WORKSPACE_BUILD_IMAGE
        self._build_steps = DEFAULT_BUILD_STEPS

    @property
    def name(self) -> str:
        return os.path.basename(self._path)

    @property
    def path(self) -> str:
        return self._path

    def _is_out_of_date(self) -> bool:
        if self._rebuild:
            return True

        # determine the most recent timestamp of the compiled workspace files
        build_path = os.path.join(self._path, 'artifacts')
        workspace_build_timestamp = get_last_modified_timestamp([build_path], 'wasm')

        # determine the timestamp of the contract source
        contract_src_paths = [f'{contract.source_path}/src' for contract in self._contracts]
        contract_source_timestamp = get_last_modified_timestamp(contract_src_paths, 'rs')

        return contract_source_timestamp > workspace_build_timestamp

    def _schedule_container(self) -> Container:
        mounts = [
            Mount('/code/target', f'workspace_{os.path.basename(self.path)}_cache'),
            Mount('/usr/local/cargo/registry', 'registry_cache'),
            Mount('/code', os.path.abspath(self.path), type='bind'),
        ]

        # get the docker client
        client = from_env()

        # start the container
        entrypoint = None if self._optimize else "/bin/sh"
        args = None if self._optimize else ["-c", " && ".join(self._build_steps)]
        return client.containers.run(
            self._build_image,
            args,
            mounts=mounts,
            entrypoint=entrypoint,
            detach=True,
        )

    def _show_logs(self):
        return self._log

def build_workspace(
    path: str,
    contracts: List[Contract],
    optimize: Optional[bool] = False,
    rebuild: Optional[bool] = False,
    log: Optional[bool] = False
):
    """
    Will attempt to build the cargo workspace including all contracts

    :param optimize: Whether to perform an optimized build
    :param rebuild: Whether to force a rebuild of the workspace
    :return:
    """
    # create all the tasks to be done
    tasks = [WorkspaceBuildTask(path, contracts, optimize, rebuild, log)]

    # run the tasks
    run_tasks(tasks)



import hashlib
import argparse
import os
import struct

from blessings import Terminal
from jenesis.contracts.build import (
    build_contracts, build_workspace, CONTRACT_BUILD_IMAGE, WORKSPACE_BUILD_IMAGE
)
from jenesis.config import Config
from jenesis.contracts.detect import detect_contracts, is_workspace
from jenesis.contracts.schema import generate_schemas, load_contract_schema
from jenesis.tasks.image import image_exists, pull_image

def _compute_init_checksum(path, contract_name):
    schema_path = os.path.join(path, "contracts", contract_name)
    schema = load_contract_schema(schema_path)

    # check for workspace-style schema
    if contract_name in schema:
        schema = schema[contract_name]

    hasher = hashlib.sha256()
    encoded_value = "" if schema is None else str(schema)

    hasher.update(struct.pack(">Q", len(encoded_value)))
    hasher.update(encoded_value.encode())
    return hasher.hexdigest()


def run(args: argparse.Namespace):

    project_path = os.getcwd()

    # check that we are actually running the command from the project root
    if not os.path.exists(os.path.join(project_path, "jenesis.toml")):
        print("Please run command from project root")
        return 1

    term = Terminal()

    contracts = detect_contracts(project_path)
    if contracts is None or len(contracts) == 0:
        print(term.red("Unable to detect any contracts"))
        return 1

    init_checksums = {contract.name: _compute_init_checksum(project_path, contract.name) for contract in contracts}

    if is_workspace(project_path):
        if not image_exists(WORKSPACE_BUILD_IMAGE):
            print(term.green("\nPulling docker image..."))
            pull_image(WORKSPACE_BUILD_IMAGE)
        print(term.green("\nBuilding cargo workspace..."))
        build_workspace(project_path, contracts, optimize=args.optimize, rebuild=args.rebuild, log=args.log)
    else:
        if not image_exists(CONTRACT_BUILD_IMAGE):
            print(term.green("\nPulling docker image..."))
            pull_image(CONTRACT_BUILD_IMAGE)
        print(term.green("\nBuilding contracts..."))
        build_contracts(contracts, batch_size=args.batch_size, optimize=args.optimize,rebuild=args.rebuild, log=args.log)

    # generate the schemas
    print(term.green("\nGenerating contract schemas..."))
    generate_schemas(contracts, batch_size=args.batch_size, rebuild=args.rebuild)

    cfg = Config.load(os.getcwd())
    for contract in contracts:
        if _compute_init_checksum(project_path, contract.name) != init_checksums[contract.name]:
            # update project file
            for (profile_name, profile) in cfg.profiles.items():
                network_name = profile.network.name
                for deployment in profile.deployments.values():
                    if deployment.contract == contract.name:
                        Config.update_project(project_path, profile_name, network_name, contract, deployment.name)
    return 0


def add_compile_command(parser):
    compile_cmd = parser.add_parser("compile")
    compile_cmd.add_argument(
        "-p",
        dest="batch_size",
        type=int,
        default=5,
        help="The limit of the number of tasks to do in parallel (default = 5)",
    )
    compile_cmd.add_argument(
        "-o",
        "--optimize",
        action="store_true",
        help="Optimize build",
    )
    compile_cmd.add_argument(
        "-r",
        "--rebuild",
        action="store_true",
        help="Force rebuild",
    )
    compile_cmd.add_argument(
        "--log",
        action="store_true",
        help="Show build logs (default)",
    )
    compile_cmd.add_argument(
        "--no-log",
        dest="log",
        action="store_false",
        help="Do not show build logs",
    )
    compile_cmd.set_defaults(handler=run, log=True)



import platform
from typing import List, Union

from jenesis.keyring.amino.codec import LocalInfo, OfflineInfo
from jenesis.keyring import test

try:
    from jenesis.keyring import macos
except Exception:
    pass


try:
    from jenesis.keyring import linux
except Exception:
    pass


def query_keychain_item(name: str, backend: str = "os") -> Union[LocalInfo, OfflineInfo]:
    if backend == "os":
        system = platform.system()
        if system == "Darwin":
            return macos.query_keychain_item(name)

        if system == "Linux":
            return linux.query_keychain_item(name)

        raise RuntimeError(f"Platform {system} not currently supported")

    if backend == "test":
        return test.query_keychain_item(name)

    raise RuntimeError(f"Keyring backend '{backend}' not currently supported")


def query_keychain_items(backend: str = "os") -> List[str]:
    if backend == "os":
        system = platform.system()
        if system == "Darwin":
            return macos.query_keychain_items()

        if system == "Linux":
            return linux.query_keychain_items()

        raise RuntimeError(f"Platform {system} not currently supported")

    if backend == "test":
        return test.query_keychain_items()

    raise RuntimeError(f"Keyring backend '{backend}' not currently supported")



import argparse
import os
import toml

from jenesis.config import Config
from jenesis.contracts.detect import detect_contracts


def run(args: argparse.Namespace):

    project_path = os.getcwd()

    # check that we are actually running the command from the project root
    if not os.path.exists(os.path.join(project_path, "jenesis.toml")):
        # pylint: disable=all
        print("Please run command from project root")
        return False

    cfg = Config.load(project_path)
    contracts = detect_contracts(project_path)

    contracts_to_update = []

    data = toml.load("jenesis.toml")

    profiles = list(cfg.profiles.keys())
    current_contracts = data["profile"][profiles[0]]["contracts"].keys()

    for contract in contracts:
        if not contract.name in current_contracts:
            contracts_to_update.append(contract)
            
    if len(contracts_to_update) == 0:
        print('Nothing to update')
        return
               
    for profile in profiles:
        for contract in contracts_to_update:
            network_name = cfg.profiles[profile].network.name
            Config.update_project(os.getcwd(), profile, network_name, contract, contract.name)
    
    print("Contracts up to date!")


def add_update_command(parser):
    update_cmd = parser.add_parser("update")
    update_cmd.set_defaults(handler=run)


# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['jenesis',
 'jenesis.cmd',
 'jenesis.cmd.add',
 'jenesis.cmd.keys',
 'jenesis.config',
 'jenesis.contracts',
 'jenesis.keyring',
 'jenesis.keyring.amino',
 'jenesis.keyring.infos',
 'jenesis.keyring.linux',
 'jenesis.keyring.macos',
 'jenesis.keyring.test',
 'jenesis.network',
 'jenesis.tasks',
 'jenesis.test']

package_data = \
{'': ['*']}

install_requires = \
['blessings>=1.7,<2.0',
 'cosmpy>=0.9.1,<0.10.0',
 'docker>=5.0.3,<6.1.0',
 'graphlib-backport>=1.0.3,<2.0.0',
 'jsonschema>=4.15.0,<5.0.0',
 'jwskate>=0.4.1,<0.6.0',
 'keyring>=23.9.0,<24.0.0',
 'makefun>=1.15.0,<2.0.0',
 'mkdocs-material>=8.3.9,<9.0.0',
 'mkdocs>=1.3.1,<2.0.0',
 'ptpython>=3.0.23,<4.0.0',
 'toml>=0.10.2,<0.11.0',
 'tqdm>=4.64.0,<5.0.0']

entry_points = \
{'console_scripts': ['jenesis = jenesis.cli:main']}

setup_kwargs = {
    'name': 'jenesis',
    'version': '0.7.0',
    'description': 'Command line tool for rapid CosmWasm-based smart contract development',
    'long_description': 'Jenesis is a command line tool for rapid contract and service development for the Fetch.ai blockchain ecosystem and other CosmWasm-enabled blockchains.\n\n# Installation\n\nInstall jenesis for Python 3.8 or newer via PyPI:\n\n```\npip install jenesis\n```\n\n# Getting started\n\n## Create a new project\n\n```\njenesis new my_project\n```\n\n## Initialize jenesis in an existing project directory\n\n```\njenesis init\n```\n\n## Compile contracts\n\n\n```\njenesis compile\n```\n',
    'author': 'Ed FitzGerald',
    'author_email': 'edward.fitzgerald@fetch.ai',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import argparse
import os
import pytest
import sys


PROJECT_PATH = os.getcwd()


def run(args: argparse.Namespace):
    pytest_args = sys.argv[sys.argv.index("test") + 1 :]
    pytest.main(pytest_args)


def add_test_command(parser):
    run_cmd = parser.add_parser("test")
    run_cmd.add_argument(
        "-p", "--profile", default=None, help="The profile to use"
    )
    run_cmd.add_argument(
        "args", nargs= "*", help="The args to pass to pytest"
    )
    run_cmd.set_defaults(handler=run)



import argparse
import os

from cosmpy.crypto.address import Address
from cosmpy.crypto.keypairs import PublicKey

from jenesis.config import Config, DEFAULT_KEYRING_BACKEND
from jenesis.keyring import query_keychain_items, query_keychain_item


def get_keyring_backend(args: argparse.Namespace) -> str:
    if args.keyring_backend:
        return args.keyring_backend
    if os.path.exists(os.path.join(os.getcwd(), "jenesis.toml")):
        return Config.load(os.getcwd()).keyring_backend
    return DEFAULT_KEYRING_BACKEND


def run_list(args: argparse.Namespace):
    key_names = query_keychain_items(get_keyring_backend(args))
    if len(key_names) == 0:
        print('No keys available')
        return

    for name in sorted(key_names):
        print(name)


def run_show(args: argparse.Namespace):
    info = query_keychain_item(args.name, get_keyring_backend(args))
    if info.algorithm != 'secp256k1':
        print(f'Unsupported key algorithm {info.algorithm}')

    public_key = PublicKey(info.public_key)
    address = Address(public_key)

    print(str(address))


def add_keys_command(parser):
    keys_cmd = parser.add_parser('keys')
    subparsers = keys_cmd.add_subparsers()

    list_cmd = subparsers.add_parser('list')
    list_cmd.add_argument('-k', '--keyring_backend', type=str, help='The keyring backend to use')
    list_cmd.set_defaults(handler=run_list)

    show_cmd = subparsers.add_parser('show')
    show_cmd.add_argument('name', help='The name of the key to show')
    show_cmd.add_argument('-k', '--keyring_backend', type=str, help='The keyring backend to use')
    show_cmd.set_defaults(handler=run_show)
